"""HTTP Navigation Manager."""

from configparser import ConfigParser
from typing import Dict, List

import logging

import random

from core.base_module import BaseModule

from core.http_manager import HttpNavigationManager
logger = logging.getLogger()


class HttpNavigationManagerHandler(BaseModule):
    """HTTP Navigation Manager."""

    def run(self, config: ConfigParser, args: Dict, data: Dict) -> None:
        """Execute Module."""

        # Check if Not have a Selected UA
        if 'web_navigation' not in data or 'user_agent' not in data['web_navigation']:
            data.update({'web_navigation': {'user_agent': self.__choose_ua(config)}})

        # Initialize Internal Modules
        HttpNavigationManager.init(data)

    def __choose_ua(self, config: ConfigParser) -> str:
        """Choose a Random UA."""

        if config:
            ua_list: List[str] = config['WEB_NAVIGATION']['useragent_list'].split('\n')
            return ua_list[random.Random().randint(1, len(ua_list) - 1)]  # nosec

        raise Exception("System Configuration Not Initialized")



"""Bitcoint Wallet Handler."""

from configparser import ConfigParser
from typing import Dict

import json
import logging
import time

import requests

from core.base_module import BaseModule
from core.temp_file import TempFileHandler

logger = logging.getLogger()


class BitcoinWalletInfoDownloader(BaseModule):
    """Download BitCoin Wallet Info."""

    def run(self, config: ConfigParser, args: Dict, data: Dict) -> None:
        """Execute Module."""

        # Get Target
        if 'bitcoin' not in data:
            data.update(
                {
                    'bitcoin': {
                        'target_wallet': args['btc_wallet']
                        }
                    }
                )

        target_wallet: str = data['bitcoin']['target_wallet']
        target_file: str = f'bitcoin_wallet/{target_wallet}_wallet.json'

        # Module Activation Rule
        if target_wallet is None or target_wallet == '':
            logger.info('\t\tTarget BTC Wallet Empty.')
            return

        # Check if Cached File Exists
        if TempFileHandler.file_exist(target_file):
            logger.info(f'\t\tTarget File "{target_file}" Alread Exists. Skipping...')
            return

        logger.info('\t\tDownloading BitCoin Wallet Info')

        # Download Wallet Data
        response: requests.Response = requests.get(
            url=config['MODULE_BitcoinWalletInfoDownloader']['wallet_info_api'].replace('{0}', target_wallet)
            )

        # Check Result
        if response.status_code == 200:

            # Save temp File
            TempFileHandler.write_file_text(target_file, response.content.decode('utf-8'))

        else:
            logger.error(response.status_code)


class BitcoinWalletTransactionsDownloader(BaseModule):
    """Download the Bitcoint Wallet Transactions."""

    def run(self, config: ConfigParser, args: Dict, data: Dict) -> None:
        """Execute Module."""

        target_wallet: str = data['bitcoin']['target_wallet']
        target_file: str = f'bitcoin_wallet/{target_wallet}_transactions.json'

        # Module Activation Rule
        if not args['btc_get_transactions']:
            return

        if target_wallet is None or target_wallet == '':
            logger.info('\t\tTarget BTC Wallet Empty.')
            return

        # Check if Cached File Exists
        if TempFileHandler.file_exist(target_file):
            logger.info(f'\t\tTarget File "{target_file}" Alread Exists. Skipping...')
            return

        logger.info('\t\tDownloading BitCoin Wallet Transactions')

        h_data: Dict = {
            'transactions': []
            }

        offset: int = 0
        while offset >= 0:

            # Download Wallet Data
            response: requests.Response = requests.get(
                url=config['MODULE_BitcoinWalletInfoDownloader']['wallet_transactions_api']
                .replace('{0}', target_wallet)
                .replace('{1}', str(offset))
                )

            # Check Result
            if response.status_code == 200:
                downloaded_data: Dict = json.loads(response.content.decode('utf-8'))

                # Update Page Indicator
                if len(downloaded_data['txs']) == 1000:
                    offset += 1000
                    time.sleep(10)

                else:
                    offset = -1

                h_data['transactions'] += downloaded_data['txs']

            else:
                logger.error(response.status_code)
                return

        # Save temp File
        TempFileHandler.write_file_text(target_file, json.dumps(h_data))



"""Helper to Operate Safely BS4."""

from typing import Optional, Tuple, Union
from bs4 import BeautifulSoup, ResultSet, Tag


class BS4Helper:
    """Helper to Operate Safely BS4."""

    def __init__(self, soup: BeautifulSoup):
        """Initialize the BS4Helper."""
        self.soup = soup

    def find_by_attribute(self, attribute_data: Tuple[str, str], target_index: int, target_property: str, null_value: Optional[str] = None, target: Optional[Union[ResultSet, Tag]] = None) -> Optional[str]:
        """Safely find Value by Attribute."""

        attr: ResultSet = self.soup.findAll(attrs={attribute_data[0]: attribute_data[1]}) if target is None else target.findAll(attrs={attribute_data[0]: attribute_data[1]})

        if attr is None or len(attr) <= 0 or len(attr) < (target_index - 1):
            return null_value

        target_tag: Tag = attr[target_index]

        if target_property == 'string':
            return self.clenup_string(target_tag.string)

        if target_property == 'text':
            return self.clenup_string(target_tag.text)

        return self.clenup_string(target_tag.attrs[target_property])

    def clenup_string(self, content: Optional[str]) -> Optional[str]:
        """Cleanup the Strings."""

        if content is None:
            return None

        h_result: str = str(content.replace('\n', ''))

        while '  ' in h_result:
            h_result = h_result.replace('  ', ' ')

        return h_result.rstrip().lstrip()



"""
OSIx - Open Source Intelligence eXplorer.

By: Th3 0bservator
"""

import sys
import importlib
import os
import types
import logging
import logging.config

from configparser import ConfigParser
from typing import Dict, List, Optional

from core.base_module import BaseModule
from core.chrome_driver_manager import ChromeDrivers
from core.dir_manager import DirectoryManagerUtils

logger = logging.getLogger()

BANNER: str = '''
OSIx - Open Source Intelligence eXplorer
Version: 0.0.3
By: Th3 0bservator
'''


class OSIx:
    """OSIx Main Module."""

    def __init__(self) -> None:
        """Module Initialization."""

        self.config: Optional[ConfigParser] = None
        self.available_modules: List[str] = []

    def main(self) -> int:
        """Application Entrypoint."""

        self.__setup_logging()

        logger.info(BANNER)

        if not self.check_python_version():
            return 1

        DirectoryManagerUtils.ensure_dir_struct('data/')
        DirectoryManagerUtils.ensure_dir_struct('data/temp/')
        DirectoryManagerUtils.ensure_dir_struct('data/export/')
        self.__load_settings()
        self.__list_modules()

        if not self.config:
            return -1

        # Load Modules
        args: Dict = {}
        data: Dict = {}
        logger.info('[*] Executing Pipeline:')
        for pipeline_item in self.config['PIPELINE']['pipeline_sequence'].split('\n'):

            if pipeline_item == '':
                continue

            logger.info(f'\t[+] {pipeline_item}')
            pipeline_item_meta: List[str] = pipeline_item.split('.')

            osix_module: types.ModuleType = importlib.import_module(f'modules.{pipeline_item_meta[0]}')
            module_instance: BaseModule = getattr(osix_module, pipeline_item_meta[1])()
            module_instance.run(
                config=self.config,
                args=args,
                data=data
                )

        # Shutdown all Drivers
        ChromeDrivers.shutdown()

        return 0

    def check_python_version(self) -> bool:
        """Check if the Current Python Version is Supported."""

        # Check Python Requirements
        major = sys.version_info[0]
        minor = sys.version_info[1]

        python_version = str(sys.version_info[0]) + "." + str(sys.version_info[1]) + "." + str(sys.version_info[2])

        if major != 3 or major == 3 and minor < 6:
            logger.fatal('This application requires at least, Python 3.7.6')
            logger.fatal(f'Current Installed Version is: {python_version}')
            return False

        return True

    def __setup_logging(self) -> None:
        """Setup Log Config."""

        logging.config.fileConfig('logging.conf')

    def __list_modules(self) -> None:
        """
        List All Available Modules.

        :return: None
        """

        # Check Modules
        logger.info('[*] Installed Modules:')
        for file in os.listdir('modules'):
            if file not in ('__init__.py', '__pycache__'):
                logger.info(f'\t{file}')

    def __load_settings(self) -> None:
        """
        Load the config.ini file into Settings Object.

        :return: None
        """

        logger.info('[*] Loading Configurations:')
        self.config = ConfigParser()
        self.config.read('config.ini')


if __name__ == '__main__':
    exit(OSIx().main())



"""Input Arguments Handler."""

from configparser import ConfigParser
from typing import Dict
from uuid import uuid4

import argparse
import logging

from core.base_module import BaseModule

logger = logging.getLogger()


class InputArgsHandler(BaseModule):
    """Module That Handle the Input Arguments."""

    __ARGS: Dict = {
        'job_name': {
            'param': '--job_name',
            'type': str,
            'action': 'store',
            'help': 'Job Name. Used to Save/Restore State File.',
            'default': uuid4().hex.upper()
            },

        'purge_temp_files': {
            'param': '--purge_temp_files',
            'type': str,
            'action': 'store_true',
            'help': 'Force Delete All Temporary Files',
            'default': False
            },

        'btc_wallet': {
            'param': '--btc_wallet',
            'type': str,
            'action': 'store',
            'help': 'BitCoin Wallet Address',
            'default': ''
            },
        'btc_get_transactions': {
            'param': '--btc_get_transactions',
            'type': str,
            'action': 'store_true',
            'help': 'Allow to Download All BitCoin Transactions from Wallet',
            'default': ''
            },
        'export_btc_transactions_as_graphml': {
            'param': '--export_btc_transactions_as_graphml',
            'type': str,
            'action': 'store_true',
            'help': 'Allow to Export the BitCoin Transactions as GraphML',
            'default': False
            },
        'export_btc_transactions_as_gephi': {
            'param': '--export_btc_transactions_as_gephi',
            'type': str,
            'action': 'store_true',
            'help': 'Allow to Export the BitCoin Transactions as Gephi File',
            'default': True
            },

        'username': {
            'param': '--username',
            'type': str,
            'action': 'store',
            'help': 'Username to Search',
            'default': ''
            },
        'username_scan': {
            'param': '--username_scan',
            'type': str,
            'action': 'store_true',
            'help': 'Allow the Executor to Scan the the Username in All Social Networks and WebSites',
            'default': False
            },
        'username_allow_nsfw_scan': {
            'param': '--username_allow_nsfw_scan',
            'type': str,
            'action': 'store_true',
            'help': 'Allow the Executor to Scan the NSFW WebSites',
            'default': False
            },
        'username_print_result': {
            'param': '--username_print_result',
            'type': str,
            'action': 'store_true',
            'help': 'Allow to Print the Result in sysout',
            'default': True
            },
        'username_show_all': {
            'param': '--username_show_all',
            'type': str,
            'action': 'store_true',
            'help': 'Allow to Print all Results, otherwise, Print Only the Founded Ones.',
            'default': False
            },
        'username_enable_dump_file': {
            'param': '--username_enable_dump_file',
            'type': str,
            'action': 'store_true',
            'help': 'Allow to Dump a Result file into data/export Folder.',
            'default': False
            },

        }

    def run(self, config: ConfigParser, args: Dict, data: Dict) -> None:
        """Execute Module."""

        parser = argparse.ArgumentParser(description='OSIx - Open Source Intelligence Explorer.')

        # Add Parameters to Arg Parser
        for arg in InputArgsHandler.__ARGS:
            spec: Dict = InputArgsHandler.__ARGS[arg]

            parser.add_argument(
                spec['param'], action=spec['action'], dest=arg,
                help=spec['help'], default=spec['default']
                )

        # Parse Args
        input_args = parser.parse_args()

        # Add to Result Args
        for arg in InputArgsHandler.__ARGS:
            args.update(
                {arg: getattr(input_args, arg)}
                )

        # Print Settings
        for key, value in args.items():
            logger.info(f'\t\t{key} = {value}')



"""Username Scanner - Based on https://github.com/sherlock-project."""
import json
from configparser import ConfigParser
from typing import Dict, List, Optional, Tuple, Union
from time import monotonic

import logging
import re
import requests

from urllib3.exceptions import InsecureRequestWarning
from requests_futures.sessions import FuturesSession
from core.base_module import BaseModule
from core.resources_manager import ResourcesFileHandler
from core.temp_file import TempFileHandler


logger = logging.getLogger()


class UsernameScanner(BaseModule):
    """Username Handler."""

    QS_CLAIMED = "Claimed"   # Username Detected
    QS_AVAILABLE = "Available"  # Username Not Detected
    QS_UNKNOWN = "Unknown"   # Error Occurred While Trying To Detect Username
    QS_ILLEGAL = "Illegal"   # Username Not Allowable For This Site

    def run(self, config: ConfigParser, args: Dict, data: Dict) -> None:
        """Execute Module."""

        # Suppress only the single warning from urllib3 needed.
        requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)  # type: ignore

        if 'username' not in data:
            data.update({'username': {'target_username': ''}})

        target_username: str = data['username']['target_username']

        if target_username == '':
            if args['username'] == '':
                logger.info('\t\tUsername Not Provided. Skipping...')
                return

            target_username = args['username']
            data['username']['target_username'] = target_username

        if not args['username_scan']:
            logger.info('\t\tUsername Scan not Enabled. Skipping...')
            return

        # Load the Targets File
        targets: Dict = json.loads(''.join(ResourcesFileHandler.read_file_text('username_scanner_targets.json')))

        # Check for NSFW Sites
        if args['username_allow_nsfw_scan']:
            logger.info('\t\tNSFW Sites Allowed.')
            targets.update(json.loads(''.join(ResourcesFileHandler.read_file_text('username_scanner_targets_nsfw.json'))))

        # Scan
        target_file: str = f'username_search/{target_username}.json'
        h_result: Optional[Dict] = None

        if not TempFileHandler.file_exist(target_file):
            h_result = self.__do_scan(
                target_username=target_username,
                target_sites=targets,
                config=config
                )
            TempFileHandler.write_file_text(target_file, json.dumps(h_result))
        else:
            logger.info(f'\t\tTarget File "{target_file}". Loading...')
            h_result = json.loads(''.join(TempFileHandler.read_file_text(target_file)))

        # Add Detected into Main Data
        self.__add_into_data(h_result, data)

        # Print Results
        self.__print_result(args, h_result)

        # Dump the Output File
        if args['username_enable_dump_file']:
            dump_file_content: str = 'site,url'

            for _key, value in h_result.items():  # type: ignore
                if 'status' in value and value['status']['status'] == UsernameScanner.QS_CLAIMED:
                    dump_file_content += f'\n"{value["status"]["site_name"]}","{value["status"]["site_url_user"]}"'

            with open(f'data/export/username_{target_username}.csv', 'w', encoding='utf-8') as file:
                file.write(dump_file_content)
                file.flush()
                file.close()

    def __print_result(self, args: Dict, h_result: Optional[Dict]) -> None:
        """Print the Result in Sysout."""

        if h_result is None or not args['username_print_result']:
            return

        for key, value in h_result.items():

            if 'status' not in value:
                continue

            if args['username_show_all'] or (not args['username_show_all'] and value['status']['status'] == UsernameScanner.QS_CLAIMED):
                logger.info(f'\t\t{key}: {value["status"]["status"]} > {value["url_user"]}')

    def __do_scan(self, target_username: str, target_sites: Dict, config: ConfigParser) -> Dict:
        """
        Execute the Scan.

        :param target_username:
        :param target_sites:
        :return:
        """

        # Create multi-threaded session for all requests.
        session: FuturesSession = self.__create_session(target_sites=target_sites)

        # Results from analysis of all sites
        results_total: Dict = self.__create_futures(
            target_username=target_username,
            target_sites=target_sites,
            config=config,
            session=session
            )

        # Parse the Results
        self.__parse_data(
            target_username=target_username,
            target_sites=target_sites,
            results_total=results_total
            )

        return results_total

    def __parse_data(self, target_username: str, target_sites: Dict, results_total: Dict) -> None:  # pylint: disable=R0914
        """
        Download and Parse WebSites Data.

        :param target_username:
        :param target_sites:
        :param results_total:
        :param args:
        :return:
        """

        # Open the file containing account links
        # Core logic: If tor requests, make them here. If multi-threaded requests, wait for responses
        for social_network, net_info in target_sites.items():

            # Retrieve results again
            results_site: Dict = results_total[social_network]

            # Retrieve other site information again
            url: str = results_site['url_user']
            status: Optional[str] = results_site['status'] if 'status' in results_site else None
            if status is not None:
                # We have already determined the user doesn't exist here
                continue

            # Get the expected error type
            error_type: str = net_info["errorType"]

            # Retrieve future and ensure it has finished
            future = net_info["request_future"]
            request_future, error_text, _expection_text = self.__get_response(
                request_future=future
                )

            if request_future is None:
                continue

            # Get response time for response of our request.
            response_time: float = float(request_future.elapsed) if request_future is not None else -1.0   # type: ignore
            http_status: int = request_future.status_code if request_future is not None else 0
            response_text: str = request_future.text if request_future is not None else ""

            if error_text is not None and error_text != '':
                result = self.__build(
                    target_username,
                    social_network,
                    url,
                    UsernameScanner.QS_UNKNOWN,
                    query_time=response_time,
                    context=error_text
                    )

            elif error_type == "message":
                result = self.__parse_single_message(net_info, request_future, response_time, social_network, target_username, url)

            elif error_type == "status_code":
                result = self.__parse_by_status_code(request_future, response_time, social_network, target_username, url)

            elif error_type == "response_url":
                result = self.__parse_by_url(request_future, response_time, social_network, target_username, url)

            else:
                # It should be impossible to ever get here...
                raise ValueError(f"Unknown Error Type '{error_type}' for "
                                 f"site '{social_network}'")

            # Save status of request
            results_site['status'] = result

            # Save results from request
            results_site['http_status'] = http_status

            try:
                results_site['response_text'] = response_text
            except:  # pylint: disable=W0702 # noqa: B001, E722, D901
                results_site['response_text'] = ''

            # Add this site's results into final dictionary with all of the other results.
            results_total[social_network] = results_site

    def __parse_by_url(self, request_future: requests.Response, response_time: float, social_network: str, target_username: str, url: str) -> Dict:  # pylint: disable=R0913
        """Parse the Result by URL."""

        # For this detection method, we have turned off the redirect.
        # So, there is no need to check the response URL: it will always
        # match the request.  Instead, we will ensure that the response
        # code indicates that the request was successful (i.e. no 404, or
        # forward to some odd redirect).
        if 200 <= request_future.status_code < 300:
            return self.__build(target_username,
                                social_network,
                                url,
                                UsernameScanner.QS_CLAIMED,
                                query_time=response_time
                                )

        return self.__build(target_username,
                            social_network,
                            url,
                            UsernameScanner.QS_AVAILABLE,
                            query_time=response_time
                            )

    def __parse_by_status_code(self, request_future: requests.Response, response_time: float, social_network: str, target_username: str, url: str) -> Dict:  # pylint: disable=R0913
        """Parse the Result by Status Code."""

        # Checks if the status code of the response is 2XX
        if not request_future.status_code >= 300 or request_future.status_code < 200:
            return self.__build(target_username,
                                social_network,
                                url,
                                UsernameScanner.QS_CLAIMED,
                                query_time=response_time
                                )

        return self.__build(target_username,
                            social_network,
                            url,
                            UsernameScanner.QS_AVAILABLE,
                            query_time=response_time
                            )

    def __parse_single_message(self, net_info: Dict, request_future: requests.Response, response_time: float, social_network: str, target_username: str, url: str) -> Dict:  # pylint: disable=R0913
        """Parse the Result by Message."""

        # error_flag True denotes no error fo
        # und in the HTML
        # error_flag False denotes error found in the HTML
        error_flag = True
        errors: Union[str, List] = net_info['errorMsg']
        # errors will hold the error message
        # it can be string or list
        # by insinstance method we can detect that
        # and handle the case for strings as normal procedure
        # and if its list we can iterate the errors
        if isinstance(errors, str):
            # Checks if the error message is in the HTML
            # if error is present we will set flag to False
            if errors in request_future.text:
                error_flag = False
        else:
            # If it's list, it will iterate all the error message
            for error in errors:
                if error in request_future.text:
                    error_flag = False
                    break

        if error_flag:
            return self.__build(target_username,
                                social_network,
                                url,
                                UsernameScanner.QS_CLAIMED,
                                query_time=response_time
                                )

        return self.__build(target_username,
                            social_network,
                            url,
                            UsernameScanner.QS_AVAILABLE,
                            query_time=response_time
                            )

    def __create_futures(self, target_username: str, target_sites: Dict, config: ConfigParser, session: FuturesSession) -> Dict:
        """Create the Request Futures."""

        # Results from analysis of all sites
        results_total: Dict = {}

        # First create futures for all requests. This allows for the requests to run in parallel
        for social_network, net_info in target_sites.items():

            # Results from analysis of this specific site
            results_site: Dict = {}

            # Record URL of main site
            results_site['url_main'] = net_info.get("urlMain")

            # A user agent is needed because some sites don't return the correct
            # information since they think that we are bots (Which we actually are...)
            headers = {
                'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.12; rv:55.0) Gecko/20100101 Firefox/55.0',
                }

            if "headers" in net_info:
                # Override/append any extra headers required by a given site.
                headers.update(net_info["headers"])

            # URL of user on site (if it exists)
            url = net_info["url"].format(target_username)

            # Don't make request if username is invalid for the site
            regex_check = net_info.get("regexCheck")
            if regex_check and re.search(regex_check, target_username) is None:
                # No need to do the check at the site: this user name is not allowed.
                results_site['status'] = self.__build(target_username,
                                                      social_network,
                                                      url,
                                                      UsernameScanner.QS_ILLEGAL)
                results_site["url_user"] = ""
                results_site['http_status'] = ""
                results_site['response_text'] = ""

            else:
                # URL of user on site (if it exists)
                results_site["url_user"] = url
                url_probe = net_info.get("urlProbe")
                if url_probe is None:
                    # Probe URL is normal one seen by people out on the web.
                    url_probe = url
                else:
                    # There is a special URL for probing existence separate
                    # from where the user profile normally can be found.
                    url_probe = url_probe.format(target_username)

                if (net_info["errorType"] == 'status_code' and net_info.get("request_head_only", True)):
                    # In most cases when we are detecting by status code,
                    # it is not necessary to get the entire body:  we can
                    # detect fine with just the HEAD response.
                    request_method = session.head
                else:
                    # Either this detect method needs the content associated
                    # with the GET response, or this specific website will
                    # not respond properly unless we request the whole page.
                    request_method = session.get

                if net_info["errorType"] == "response_url":
                    # Site forwards request to a different URL if username not
                    # found.  Disallow the redirect so we can capture the
                    # http status from the original URL request.
                    allow_redirects = False
                else:
                    # Allow whatever redirect that the site wants to do.
                    # The final result of the request will be what is available.
                    allow_redirects = True

                # This future starts running the request in a new thread, doesn't block the main thread
                future = self.__get_future(allow_redirects, config, headers, request_method, url_probe)

                # Store future in data for access later
                net_info["request_future"] = future

            # Add this site's results into final dictionary with all of the other results.
            results_total[social_network] = results_site

        return results_total

    def __get_future(self, allow_redirects: bool, config: ConfigParser, headers: Dict, request_method: FuturesSession, url_probe: str) -> FuturesSession:  # pylint: disable=R0913
        """Get the Future."""

        return request_method(
            url=url_probe,
            headers=headers,
            allow_redirects=allow_redirects,
            timeout=int(config['MODULE_UsernameScanner']['connection_timeout_sec'])
            )

    def __create_session(self, target_sites: Dict) -> FuturesSession:
        """Create the HTTP Session."""

        # Normal requests
        underlying_session = requests.session()
        underlying_session.verify = False

        # Limit number of workers to 20.
        if len(target_sites) >= 20:
            max_workers = 20
        else:
            max_workers = len(target_sites)

        logger.info(f'\t\tStarting Scan with {max_workers} Workers.')

        # Create multi-threaded session for all requests.
        return SherlockFuturesSession(
            max_workers=max_workers,
            session=underlying_session
            )

    def __get_response(self, request_future: FuturesSession) -> Tuple[Optional[requests.Response], str, str]:

        # Default for Response object if some failure occurs.
        response: Optional[requests.Response] = None

        error_context: str = 'General Unknown Error'
        expected_text: str = ''

        try:
            response = request_future.result()

            if response is not None and response.status_code:
                error_context = ''

        except requests.exceptions.HTTPError as errh:
            error_context = "HTTP Error"
            expected_text = str(errh)

        except requests.exceptions.ProxyError as errp:
            error_context = "Proxy Error"
            expected_text = str(errp)

        except requests.exceptions.ConnectionError as errc:
            error_context = "Error Connecting"
            expected_text = str(errc)

        except requests.exceptions.Timeout as errt:
            error_context = "Timeout Error"
            expected_text = str(errt)

        except requests.exceptions.RequestException as err:
            error_context = "Unknown Error"
            expected_text = str(err)

        return response, error_context, expected_text

    def __get_future_result(self, request_future: FuturesSession) -> requests.Response:
        """The the Future Result."""

        return request_future.result()  # type: ignore

    def __build(self, username: str, site_name: str, site_url_user: str, status: str, query_time: Optional[float] = None, context: str = None) -> Dict:  # pylint: disable=R0913
        """
        Create Query Result Object.

        Contains information about a specific method of detecting usernames on
        a given type of web sites.

        :param username: String indicating username that query result was about
        :param site_name: String which identifies site
        :param site_url_user: String containing URL for username on site
        :param status: Indicate the status of the query
        :param query_time: Time (in seconds) required to perform query
        :param context: String indicating any additional context about the query.  For example, if there was an error, this might indicate the type of error that occurred
        :return:
        """

        return {
            'username': username,
            'site_name': site_name,
            'site_url_user': site_url_user,
            'status': status,
            'query_time': query_time,
            'context': context
            }

    def __add_into_data(self, h_result: Optional[Dict], data: Dict) -> None:
        """Add the Detected Accounts into Data."""

        if h_result is None:
            return

        # Add all Usernames and Accounts to Main Data
        if 'detected' not in data['username']:
            data['username']['detected'] = {}

        for network, network_data in h_result.items():
            data['username']['detected'][network] = network_data

            if 'response_text' in data['username']['detected'][network]:
                del data['username']['detected'][network]['response_text']


class SherlockFuturesSession(FuturesSession):  # type: ignore
    """Sherlock Future Session."""

    def request(self, method, url, hooks={}, *args, **kwargs):  # type: ignore # pylint: disable=R0913, W1113, W0102, W0221 # noqa: B006
        """
        Request.

        This extends the FuturesSession request method to calculate a response
        time metric to each request.
        It is taken (almost) directly from the following StackOverflow answer:
        https://github.com/ross/requests-futures#working-in-the-background

        :param method: String containing method desired for request.
        :param url: String containing URL for request.
        :param hooks: Dictionary containing hooks to execute after request finishes.
        :param args: Arguments.
        :param kwargs: Keyword arguments.
        :return:
        """

        # Record the start time for the request.
        start = monotonic()

        def response_time(resp, *args, **kwargs):  # type: ignore # pylint:disable=W0613
            """Response Time Hook."""

            resp.elapsed = monotonic() - start

            return resp

        # Install hook to execute when response completes.
        # Make sure that the time measurement hook is first, so we will not
        # track any later hook's execution time.
        try:
            if isinstance(hooks['response'], list):
                hooks['response'].insert(0, response_time)
            elif isinstance(hooks['response'], tuple):
                # Convert tuple to list and insert time measurement hook first.
                hooks['response'] = list(hooks['response'])
                hooks['response'].insert(0, response_time)  # type: ignore
            else:
                # Must have previously contained a single hook function,
                # so convert to list.
                hooks['response'] = [response_time, hooks['response']]
        except KeyError:
            # No response hook was already defined, so install it ourselves.
            hooks['response'] = [response_time]

        return super(SherlockFuturesSession, self).request(method, url, hooks=hooks, *args, **kwargs)  # pylint: disable=R1725



"""Directory Manager."""

import os


class DirectoryManagerUtils:
    """Directory Manager."""

    @staticmethod
    def ensure_dir_struct(path: str) -> None:
        """
        Ensur That Directory Exists.

        :param path:
        :return:
        """

        target_path: str = os.path.abspath(os.path.join(os.getcwd(), path))

        if not os.path.exists(target_path):
            os.mkdir(target_path)



"""Chrome Driver Manager."""

import time
import logging
import zipfile

from os import path
from sys import platform

from typing import Dict, Optional

from configparser import ConfigParser
from selenium import webdriver
from selenium.common.exceptions import WebDriverException

from core.http_manager import HttpNavigationManager

logger = logging.getLogger()


class FacebookDriver:
    """Facebook Driver Manager."""

    def __init__(self, config: ConfigParser) -> None:
        """Initialize Module."""

        try:
            kwargs: Dict = self.__get_webdriver_kwargs(config=config)
            self.browser = webdriver.Chrome(**kwargs)

        except WebDriverException as ex:
            logger.fatal('[!!!] ERROR - You Probably don\'t Have the Chrome Driver Installed. Please, check "https://sites.google.com/a/chromium.org/chromedriver/home" to Install.')
            raise ex

        except Exception as ex:  # noqa: B902 - Fallback 1 - Try the SYSTEM ChromeWebDriver
            logger.error('', ex)  # pylint: disable=E1205
            self.browser = self.__init_fallback_system()

        self.login(config)

    def __init_fallback_system(self) -> webdriver.Chrome:
        """
        Fallback WebDriver Initialization to System Version.

        :return:
        """

        try:
            logger.warning('\t\tCHROME WEB DRIVER - Initialization Fail. Trying Fallback 1 - System Version')
            return webdriver.Chrome()

        except WebDriverException as ex:
            logger.fatal('[!!!] ERROR - You Probably don\'t Have the Chrome Driver Installed. Please, check "https://sites.google.com/a/chromium.org/chromedriver/home" to Install.')
            raise ex

    def __get_webdriver_kwargs(self, config: ConfigParser) -> Dict:
        """
        Return the WebDriver Initialization KWARGS.

        :return:
        """

        is_win32: bool = (platform == 'win32')
        executable_path: str = path.join("assets", 'chromedriver.exe' if platform == 'win32' else 'chromedriver')
        kwargs: Dict = {}

        # Check if Assets Exists, Otherwise, Download Them.
        if not path.exists(executable_path):
            downloadable_file_path: str = path.join('assets', 'chrome_driver.zip')

            logger.info('\t[+] Downloading Chrome Web Driver for %s', 'Windows' if is_win32 else 'Linux')
            HttpNavigationManager.download_file(
                config['WEB_NAVIGATION']['chrome_webdriver_download_uri_win32'] if is_win32 else
                config['WEB_NAVIGATION']['chrome_webdriver_download_uri_linux64'],
                downloadable_file_path
                )

            logger.info('\t\tUnzipping')
            with zipfile.ZipFile(downloadable_file_path, 'r') as zip_ref:
                zip_ref.extractall(path.dirname(executable_path))
            logger.info('\t\tDone')

        kwargs['executable_path'] = executable_path

        return kwargs

    def get_browser(self) -> webdriver.Chrome:
        """Return the WebDriver."""
        return self.browser

    def login(self, config: ConfigParser) -> None:
        """Login on Facebook."""

        # Start Navigation
        self.browser.maximize_window()
        self.browser.get('https://www.facebook.com/')

        # Auto Login
        time.sleep(1)
        self.browser.find_element_by_id("email").send_keys(config['FACEBOOK']['email'])

        time.sleep(1)
        self.browser.find_element_by_id("pass").send_keys(config['FACEBOOK']['password'])

        time.sleep(1)
        self.browser.find_elements_by_name("login")[0].click()

        time.sleep(2)
        self.browser.minimize_window()

    def shutdown(self) -> None:
        """Shutdown Web Driver."""
        self.browser.close()


class ChromeDrivers:
    """Drivers Singleton Managment."""

    __FACEBOOK_DRIVER: Optional[FacebookDriver] = None

    @staticmethod
    def shutdown() -> None:
        """Shutdown All Drivers."""

        if ChromeDrivers.__FACEBOOK_DRIVER is not None:
            ChromeDrivers.__FACEBOOK_DRIVER.shutdown()

    @staticmethod
    def get_fb_webdriver(config: ConfigParser) -> webdriver.Chrome:
        """Return the FB WebDriver."""

        if ChromeDrivers.__FACEBOOK_DRIVER is None:
            ChromeDrivers.__FACEBOOK_DRIVER = FacebookDriver(config)

        return ChromeDrivers.__FACEBOOK_DRIVER.get_browser()



"""Bitcoin Wallet Graph Generator."""

from configparser import ConfigParser
from typing import Dict, List

import json
import logging
import time

import networkx as nx

from core.base_module import BaseModule
from core.temp_file import TempFileHandler

logger = logging.getLogger()


class BitcoinWalletGraphGenerator(BaseModule):
    """Bitcoint Wallet Graph Generator."""

    def run(self, config: ConfigParser, args: Dict, data: Dict) -> None:
        """Execute Module."""

        target_wallet: str = data['bitcoin']['target_wallet']
        target_file: str = f'bitcoin_wallet/{target_wallet}_transactions.json'

        if target_wallet is None or target_wallet == '':
            logger.info('\t\tTarget BTC Wallet Empty.')
            return

        # Check if Cached File Exists
        if not TempFileHandler.file_exist(target_file):
            logger.info(f'\t\tTarget File "{target_file}" Do not Exists. Skipping...')
            return

        logger.info('\t\tGenerating BTC Graph')

        # Load File
        json_data: Dict = json.loads(''.join(TempFileHandler.read_file_text(target_file)))

        # Execute Module
        self.__execute(
            target_wallet=target_wallet,
            json_data=json_data,
            args=args
            )

    def __execute(self, target_wallet: str, json_data: Dict, args: Dict) -> None:
        """
        Execute the Graph Engine.

        :param target_wallet:
        :param target_file:
        :param json_data:
        :return:
        """

        # Initialize Graph Engine
        all_graph: nx.Graph = nx.DiGraph()
        in_graph: nx.Graph = nx.DiGraph()
        out_graph: nx.Graph = nx.DiGraph()

        all_graph.add_node(target_wallet, weight=2)
        in_graph.add_node(target_wallet, weight=2)
        out_graph.add_node(target_wallet, weight=2)

        # Generate Nodes Graph Nodes List
        all_input_nodes: List = []
        all_output_nodes: List = []
        for single_trans in json_data['transactions']:

            # Process Input
            inputs: List = single_trans['inputs']

            for single_input in inputs:
                if 'prev_out' in single_input and single_input['prev_out']['addr'] not in all_input_nodes:
                    all_input_nodes.append(single_input['prev_out']['addr'])

            # Process Output
            outputs: List = single_trans['out']

            for single_output in outputs:
                if single_output['addr'] not in all_output_nodes:
                    all_output_nodes.append(single_output['addr'])

        # Process Input Nodes
        for in_node in all_input_nodes:
            all_graph.add_node(in_node)
            all_graph.add_edge(in_node, target_wallet, key=int(time.clock() * 1000000))

            in_graph.add_node(in_node)
            in_graph.add_edge(in_node, target_wallet, key=int(time.clock() * 1000000))

        # Process Output Nodes
        for out_node in all_output_nodes:
            all_graph.add_node(out_node)
            all_graph.add_edge(target_wallet, out_node, key=int(time.clock() * 1000000))

            out_graph.add_node(out_node)
            out_graph.add_edge(target_wallet, out_node, key=int(time.clock() * 1000000))

        # Set Node Properties
        if args['export_btc_transactions_as_gephi']:
            all_graph.nodes[target_wallet]['viz'] = {'color': {'r': 255, 'g': 0, 'b': 0, 'a': 0}}
            out_graph.nodes[target_wallet]['viz'] = {'color': {'r': 255, 'g': 0, 'b': 0, 'a': 0}}
            in_graph.nodes[target_wallet]['viz'] = {'color': {'r': 255, 'g': 0, 'b': 0, 'a': 0}}

            nx.readwrite.write_gexf(all_graph, f'data/export/btc_wallet_all_{target_wallet}.gexf')
            nx.readwrite.write_gexf(out_graph, f'data/export/btc_wallet_outputs_{target_wallet}.gexf')
            nx.readwrite.write_gexf(in_graph, f'data/export/btc_wallet_inputs_{target_wallet}.gexf')

        elif args['export_btc_transactions_as_graphml']:
            nx.readwrite.write_graphml_xml(all_graph, f'data/export/btc_wallet_all_{target_wallet}.graphml')
            nx.readwrite.write_graphml_xml(out_graph, f'data/export/btc_wallet_outputs_{target_wallet}.graphml')
            nx.readwrite.write_graphml_xml(in_graph, f'data/export/btc_wallet_inputs_{target_wallet}.graphml')

        # Log
        logger.info('\t\t3 Exported Files at data/export')



"""OSIx Base Module."""

import abc
from configparser import ConfigParser
from typing import Dict


class BaseModule:
    """Base Module Declaration."""

    @abc.abstractmethod
    def run(self, config: ConfigParser, args: Dict, data: Dict) -> None:
        """
        Base Run Description.

        :return: None
        """



"""Base Module for Simple Username Data Diggers."""

import abc
import logging
import hashlib
from configparser import ConfigParser
from typing import Dict, Optional, Tuple

from core.base_module import BaseModule
from core.http_manager import HttpNavigationManager
from core.temp_file import TempFileHandler

logger = logging.getLogger()


class SimpleUsernameDataDigger(BaseModule):
    """Simple Username Data Digger."""

    @abc.abstractmethod
    def run(self, config: ConfigParser, args: Dict, data: Dict) -> None:
        """
        Base Run Description.

        :return: None
        """

    def _can_activate(self, data: Dict) -> Tuple[bool, Optional[str]]:
        """
        Check if Can Activate the Digger.

        :param data: Data Dict
        :return: A Boolean Indicator about the Activation and the Target Username
        """

        if 'username' not in data:
            logger.info('\t\tUsername Not Provided. Skipping...')
            return False, None

        target_username: str = data['username']['target_username']

        if target_username == '':
            logger.info('\t\tUsername Not Provided. Skipping...')
            return False, None

        return True, target_username

    def _download_text(self, url: str) -> str:
        """Download Remote Data."""

        # Check Temp File
        targer_file_name: str = hashlib.md5(url.encode('utf-8')).hexdigest()  # nosec

        # Load Temporary Files
        target_file_path: str = f'username_search/{targer_file_name}.json'

        if TempFileHandler.file_exist(target_file_path):
            return ''.join(TempFileHandler.read_file_text(target_file_path))

        # File Not Exists, Download and Save
        content: str = HttpNavigationManager.get(url)

        # Save Temp File
        TempFileHandler.write_file_text(target_file_path, content)

        # Return
        return content



"""HTTP Navigation Manager."""

from typing import Dict, Optional

import requests


class HttpNavigationManager:
    """Http Navigation Manager."""

    __INSTANCE: Optional[requests.Session] = None

    @staticmethod
    def init(data: Dict) -> None:
        """Initialize Navigation Manager."""

        HttpNavigationManager.__INSTANCE = requests.Session()
        HttpNavigationManager.__INSTANCE.headers.update(
            {'User-Agent': data['web_navigation']['user_agent']}
            )

    @staticmethod
    def get(uri: str) -> str:
        """
        Realize a GET on Remote URI.

        :param uri: Target URI
        :return:
        """

        if not HttpNavigationManager.__INSTANCE:
            raise Exception("HttpNavigationManager not Initialized.")

        return HttpNavigationManager.__INSTANCE.get(uri).text

    @staticmethod
    def download_file(uri: str, target_path: str) -> None:
        """
        Download a Single File from Web.

        :param uri: Source File.
        :param target_path: Target File.
        :return:
        """

        if not HttpNavigationManager.__INSTANCE:
            raise Exception("HttpNavigationManager not Initialized.")

        with open(target_path, 'wb') as file:
            file.write(
                HttpNavigationManager.__INSTANCE.get(uri).content
                )

            file.flush()
            file.close()



"""Resources File Handle."""

import os
from typing import List


class ResourcesFileHandler:
    """Resources File Hander."""

    @staticmethod
    def file_exist(path: str) -> bool:
        """
        Return if a File Exists.

        :param path: File Path
        :return:
        """

        return os.path.exists(f'resources/{path}')

    @staticmethod
    def read_file_text(path: str) -> List[str]:
        """
        Read All File Content.

        :param path: File Path
        :return: File Content
        """

        with open(f'resources/{path}', 'r', encoding='utf-8') as file:
            return file.readlines()



"""Temporary Files Manager."""

import os
import logging

from configparser import ConfigParser
from datetime import datetime
from typing import Dict

import pytz

from core.base_module import BaseModule
from core.temp_file import TempFileHandler

logger = logging.getLogger()


class TempFileManager(BaseModule):
    """Temporary File Manager."""

    def run(self, config: ConfigParser, args: Dict, data: Dict) -> None:
        """Execute Module."""

        # Check Folder Structure
        self.__ensure_folder_structure(config=config)

        if args['purge_temp_files']:
            self.__purge_all(config=config)

        else:
            self.__check_by_age(config=config)

    def __ensure_folder_structure(self, config: ConfigParser) -> None:
        """Ensure the Folders Structure."""

        for temp_file_info in config['TEMP_FILES']:
            temp_def: str = config['TEMP_FILES'][temp_file_info]
            TempFileHandler.ensure_dir_struct(temp_def.split(';')[0])

    def __check_folder_items_age(self, path: str, max_age_seconds: int) -> None:
        """
        Check Folder Items Age.

        :param path: Path to Check
        :param max_age_seconds: Max Age in Seconds
        :return: None
        """
        logger.debug(f'\t\tChecking Age {path} for {max_age_seconds} seconds')
        for file in os.scandir(path):
            file_age: int = int(os.stat(file).st_ctime)

            if (int(datetime.now(tz=pytz.UTC).timestamp()) - file_age) > max_age_seconds:
                os.remove(file)

    def __purge_folder(self, path: str) -> None:
        """
        Purge a Single Folder.

        :param path: Folder Path
        :return:
        """

        logger.debug(f'\t\tPurging {path}')

        for file in os.scandir(f'data/temp/{path}'):
            os.remove(f'data/temp/{path}/{file.name}')

    def __purge_all(self, config: ConfigParser) -> None:
        """Purge all Folders."""

        for temp_file_info in config['TEMP_FILES']:
            temp_def: str = config['TEMP_FILES'][temp_file_info]
            self.__purge_folder(temp_def.split(';')[0])

    def __check_by_age(self, config: ConfigParser) -> None:
        """Check Folders file By Age."""

        for temp_file_info in config['TEMP_FILES']:
            temp_def: str = config['TEMP_FILES'][temp_file_info]
            folder_name: str = temp_def.split(';')[0]
            age: int = int(temp_def.split(';')[1])
            self.__check_folder_items_age(f'data/temp/{folder_name}', age)



"""Constants File."""

from typing import Dict

SINGLE_FB_PROFILE_DATA_ITEM: Dict = {
    'id': '',  # Facebook ID
    'friends': [],  # List of SINGLE_FB_FRIEND_DATA_ITEM
    'family': []
    }

SINGLE_INSTAGRAM_PROFILE_DATA_ITEM: Dict = {
    'fullname': '',
    'profile_pic': '',
    'id': '',  # Instagram ID
    'followers': [],  # List of SINGLE_INSTAGRAM_FRIEND_DATA_ITEM
    'following': []  # List of SINGLE_INSTAGRAM_FRIEND_DATA_ITEM
    }

SINGLE_FB_FRIEND_DATA_ITEM: Dict = {
    'name': '',
    'id': '',
    'path': '',
    'profile_pic': ''
    }

SINGLE_INSTAGRAM_FRIEND_DATA_ITEM: Dict = {
    'name': '',
    'id': '',
    'path': '',
    'profile_pic': ''
    }



"""Temp File Handle."""

import os
from typing import List


class TempFileHandler:
    """Temporary File Hander."""

    ROOT_PATH: str = os.path.join(os.getcwd())

    @staticmethod
    def ensure_dir_struct(path: str) -> None:
        """
        Ensur That Directory Exists.

        :param path:
        :return:
        """

        if not os.path.exists(os.path.abspath(os.path.join(TempFileHandler.ROOT_PATH, 'data/temp/'))):
            os.mkdir(os.path.abspath(os.path.join(TempFileHandler.ROOT_PATH, 'data/temp/')))

        if not os.path.exists(os.path.abspath(os.path.join(TempFileHandler.ROOT_PATH, f'data/temp/{path}'))):
            os.mkdir(os.path.abspath(os.path.join(TempFileHandler.ROOT_PATH, f'data/temp/{path}')))

    @staticmethod
    def file_exist(path: str) -> bool:
        """
        Return if a File Exists.

        :param path: File Path
        :return:
        """

        return os.path.exists(os.path.abspath(os.path.join(TempFileHandler.ROOT_PATH, f'data/temp/{path}')))

    @staticmethod
    def read_file_text(path: str) -> List[str]:
        """
        Read All File Content.

        :param path: File Path
        :return: File Content
        """

        with open(os.path.abspath(os.path.join(TempFileHandler.ROOT_PATH, f'data/temp/{path}')), 'r', encoding='utf-8') as file:
            return file.readlines()

    @staticmethod
    def write_file_text(path: str, content: str) -> None:
        """
        Write Text Content into File.

        :param path: File Path
        :return: None
        """

        with open(os.path.abspath(os.path.join(TempFileHandler.ROOT_PATH, f'data/temp/{path}')), 'w', encoding='utf-8') as file:
            file.write(content)
            file.flush()
            file.close()



"""GitHub Username Data Digger - Digger Downloaded Data for Usefull Information."""
import logging

from configparser import ConfigParser
from typing import Dict, List

from bs4 import BeautifulSoup, ResultSet, Tag

from core.base_username_data_digger import SimpleUsernameDataDigger
from core.bs4_helper import BS4Helper

logger = logging.getLogger()


class GithubUsernameDataDigger(SimpleUsernameDataDigger):
    """Github Username Data Digger."""

    def run(self, config: ConfigParser, args: Dict, data: Dict) -> None:
        """Execute Module."""

        # Check the Activation and Get The Social Network Data
        can_activate, username = self._can_activate(data=data)

        if not can_activate or username is None:
            return

        # Download the GitHub Profile Page Data
        base_url = config['MODULE_GithubUsernameDataDigger']['profile_url'].replace('{0}', username)
        github_data_profile: str = self._download_text(base_url)

        # Load HTML
        bs4_helper_profile: BS4Helper = BS4Helper(soup=BeautifulSoup(github_data_profile, 'html.parser'))

        # Get all Data
        h_result: Dict = {
            'profile_pic': bs4_helper_profile.find_by_attribute(attribute_data=('alt', 'Avatar'), target_index=0, target_property='src', null_value=None),
            'fullname': bs4_helper_profile.find_by_attribute(attribute_data=('itemprop', 'name'), target_index=0, target_property='string', null_value=None),
            'nickname': bs4_helper_profile.find_by_attribute(attribute_data=('itemprop', 'additionalName'), target_index=0, target_property='string', null_value=None),
            'bio': bs4_helper_profile.find_by_attribute(attribute_data=('class', 'user-profile-bio'), target_index=0, target_property='string', null_value=None),
            'location': bs4_helper_profile.find_by_attribute(attribute_data=('itemprop', 'homeLocation'), target_index=0, target_property='text', null_value=None),
            'website': bs4_helper_profile.find_by_attribute(attribute_data=('itemprop', 'url'), target_index=0, target_property='text', null_value=None),
            'account_id': bs4_helper_profile.find_by_attribute(attribute_data=('role', 'search'), target_index=0, target_property='data-scope-id', null_value=None),
            'repos': self.__get_repos(base_url),
            'followers': self.__get_followers(base_url),
            'following': self.__get_following(base_url)
            }

        # Check if Found Anything
        if h_result['account_id'] is None:
            logger.info('\t\tUsername Not Found.')
            return

        # Add to Data
        data['github'] = h_result

        # Dump the Output File
        if args['username_print_result']:
            logger.info(f'\t\tFull Name.......: {h_result["fullname"]}')
            logger.info(f'\t\tNick Name.......: {h_result["nickname"]}')
            logger.info(f'\t\tBio.............: {h_result["bio"]}')
            logger.info(f'\t\tLocation........: {h_result["location"]}')
            logger.info(f'\t\tWebSite.........: {h_result["website"]}')
            logger.info(f'\t\tAccount ID......: {h_result["account_id"]}')
            logger.info(f'\t\tProfile Picture.: {h_result["profile_pic"]}')

            logger.info(f'\t\tRepos...........: {len(h_result["repos"])}')
            _ = [logger.info(f'\t\t\t{item["name"]} ({item["description"]}) at https://github.com{item["url"]}') for item in h_result['repos']]  # type: ignore

            logger.info(f'\t\tFollowers.......: {len(h_result["followers"])}')
            _ = [logger.info(f'\t\t\t{item["username"]} ({item["name"]}) at https://github.com{item["url"]}') for item in h_result['followers']]  # type: ignore

            logger.info(f'\t\tFollowing.......: {len(h_result["following"])}')
            _ = [logger.info(f'\t\t\t{item["username"]} ({item["name"]}) at https://github.com{item["url"]}') for item in h_result['following']]  # type: ignore

    def __get_following(self, base_url: str) -> List[Dict]:
        """Get the Following."""

        bs4_helper_following: BS4Helper = BS4Helper(
            soup=BeautifulSoup(self._download_text(base_url + '?tab=following'), 'html.parser')
            )

        all_following: List = []

        following_list_node: ResultSet = bs4_helper_following.soup.find_all(attrs={'class': 'd-inline-block no-underline mb-1'})

        if len(following_list_node) > 0:
            for item in following_list_node:
                if isinstance(item, Tag):

                    all_following.append(
                        {
                            'name': bs4_helper_following.find_by_attribute(attribute_data=('class', 'f4 Link--primary'), target_property='text', target_index=0, target=item),
                            'username': bs4_helper_following.find_by_attribute(attribute_data=('class', 'Link--secondary'), target_property='text', target_index=0, target=item),
                            'url': item.attrs['href']
                            }
                        )

        return all_following

    def __get_followers(self, base_url: str) -> List[Dict]:
        """Get User Followers."""

        bs4_helper_followers: BS4Helper = BS4Helper(
            soup=BeautifulSoup(self._download_text(base_url + '?tab=followers'), 'html.parser')
            )

        all_followers: List = []

        followers_list_node: ResultSet = bs4_helper_followers.soup.find_all(attrs={'class': 'd-inline-block no-underline mb-1'})

        if len(followers_list_node) > 0:
            for item in followers_list_node:
                if isinstance(item, Tag):
                    all_followers.append(
                        {
                            'name': bs4_helper_followers.find_by_attribute(attribute_data=('class', 'f4 Link--primary'), target_property='text', target_index=0, target=item),
                            'username': bs4_helper_followers.find_by_attribute(attribute_data=('class', 'Link--secondary'), target_property='text', target_index=0, target=item),
                            'url': item.attrs['href']
                            }
                        )

        return all_followers

    def __get_repos(self, base_url: str) -> List[Dict]:
        """Get Repo List."""

        bs4_helper_repos: BS4Helper = BS4Helper(
            soup=BeautifulSoup(self._download_text(base_url + '?tab=repositories'), 'html.parser')
            )

        all_repos: List = []

        repo_list_node: ResultSet = bs4_helper_repos.soup.find_all(attrs={'data-filterable-for': 'your-repos-filter'})

        if len(repo_list_node) > 0:
            for item in repo_list_node[0].children:
                if isinstance(item, Tag):
                    all_repos.append(
                        {
                            'name': bs4_helper_repos.find_by_attribute(attribute_data=('itemprop', 'name codeRepository'), target_property='text', target_index=0, target=item),
                            'description': bs4_helper_repos.find_by_attribute(attribute_data=('itemprop', 'description'), target_property='text', target_index=0, target=item),
                            'url': bs4_helper_repos.find_by_attribute(attribute_data=('itemprop', 'name codeRepository'), target_property='href', target_index=0, target=item),
                            'type': 'fork' if 'fork' in item.attrs['class'] else 'owner'
                            }
                        )

        return all_repos



import io
import os

from setuptools import find_packages, setup

# Package meta-data.
NAME = 'OSIx'
DESCRIPTION = 'Open Source Intelligence eXplorer.'
URL = 'https://github.com/guibacellar/OSIx/'
AUTHOR = 'Th3 0bservator'
EMAIL = 'th30bservator@gmail.com'
REQUIRES_PYTHON = '>=3.7.6'
VERSION = '0.4.2'

# What packages are optional?
EXTRAS = {
}

here = os.path.abspath(os.path.dirname(__file__))

# What packages are required for this module to be executed?
try:
    with io.open(os.path.join(here, 'requirements.txt'), encoding='utf-8') as f:
        requirements_data = f.read()
except FileNotFoundError:
    requirements_data = ''
REQUIRED = [req for req in requirements_data.split('\n') if '==' in req and 'tox' not in req and 'pytest' not in req]
print(REQUIRED)

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README_PYPI.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Where the magic happens:
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    keyword=[
        "CyberSecurity",
        "Investigation",
        "OSINT",
        "OpenSourceIntelligence",
        "Tool"
    ],
    packages=find_packages(
        exclude=["tests", "*.tests", "*.tests.*", "tests.*", "venv.*"],
        where="."
    ),
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license='Apache-2.0',
    classifiers=[
        'Environment :: Console',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Security',
        'Topic :: Utilities'
    ],
    project_urls={
        'Bug Tracker': 'https://github.com/guibacellar/OSIx/issues',
        'Documentation': 'https://github.com/guibacellar/OSIx/blob/develop/README.md',
        'Source Code': 'https://github.com/guibacellar/OSIx/'
    }
)



"""State File Handler."""

from configparser import ConfigParser
from typing import Dict

import json
import logging

from core.base_module import BaseModule
from core.temp_file import TempFileHandler

logger = logging.getLogger()


class LoadStateFileHandler(BaseModule):
    """Module that Loads Previous Created State File."""

    def run(self, config: ConfigParser, args: Dict, data: Dict) -> None:
        """Execute Module."""

        state_file_name: str = config['MODULE_LoadStateFileHandler']['file_name'].replace("{0}", args['job_name'])

        if TempFileHandler.file_exist(state_file_name):
            data.update(
                json.loads(''.join(TempFileHandler.read_file_text(state_file_name)))
                )
            logger.debug("\t\tState File Loaded.")


class SaveStateFileHandler(BaseModule):
    """Module that Save a New State File."""

    def run(self, config: ConfigParser, args: Dict, data: Dict) -> None:
        """Execute Module."""

        state_file_name: str = config['MODULE_SaveStateFileHandler']['file_name'].replace("{0}", args['job_name'])

        TempFileHandler.write_file_text(
            state_file_name,
            json.dumps(data)
            )
